# -*- coding: utf-8 -*-
"""
Created on Sun Jun 22 11:29:58 2025

@author: Thomas Lee
Rice University
Department of Earth, Environmental, and Planetary Sciences
Email: tl165@rice.edu
"""
all = ['database']
DATA_FILES = ["AviList-v2025-11Jun-extended.xlsx","AviListDataBase.pickle"]